use crate::App;
use crate::gobject_models::GSidebarSelection;
use gdk4::Rectangle;
use glib::{Properties, SignalHandlerId, subclass::*};
use gtk4::{CompositeTemplate, ListBox, ListBoxRow, PopoverMenu, Widget, prelude::*, subclass::prelude::*};
use libadwaita::{Bin, subclass::prelude::*};
use once_cell::sync::Lazy;
use std::cell::Cell;

mod imp {
    use super::*;

    #[derive(Debug, CompositeTemplate, Default, Properties)]
    #[properties(wrapper_type = super::AllArticlesRow)]
    #[template(file = "data/resources/ui_templates/sidebar/all_articles_row.blp")]
    pub struct AllArticlesRow {
        #[template_child]
        pub list: TemplateChild<ListBox>,
        #[template_child]
        pub row: TemplateChild<ListBoxRow>,
        #[template_child]
        pub popover: TemplateChild<PopoverMenu>,

        #[property(get, set, name = "item-count")]
        pub item_count: Cell<u32>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for AllArticlesRow {
        const NAME: &'static str = "AllArticlesRow";
        type ParentType = Bin;
        type Type = super::AllArticlesRow;

        fn class_init(klass: &mut Self::Class) {
            klass.bind_template();
            klass.bind_template_callbacks();
        }

        fn instance_init(obj: &InitializingObject<Self>) {
            obj.init_template();
        }
    }

    #[glib::derived_properties]
    impl ObjectImpl for AllArticlesRow {
        fn signals() -> &'static [Signal] {
            static SIGNALS: Lazy<Vec<Signal>> = Lazy::new(|| {
                vec![
                    Signal::builder("activated")
                        .param_types([super::AllArticlesRow::static_type()])
                        .build(),
                    Signal::builder("selection-changed")
                        .param_types([super::AllArticlesRow::static_type(), GSidebarSelection::static_type()])
                        .build(),
                ]
            });

            SIGNALS.as_ref()
        }
    }

    impl WidgetImpl for AllArticlesRow {}

    impl BinImpl for AllArticlesRow {}

    #[gtk4::template_callbacks]
    impl AllArticlesRow {
        #[template_callback]
        fn on_clicked(&self, times: i32, x: f64, y: f64) {
            if times != 1 {
                return;
            }

            if App::default().is_offline() {
                return;
            }

            let position = Rectangle::new(x as i32, y as i32, 0, 0);
            self.popover.set_pointing_to(Some(&position));
            self.popover.popup();
        }

        #[template_callback]
        fn on_pressed(&self, x: f64, y: f64) {
            if App::default().is_offline() {
                return;
            }

            let position = Rectangle::new(x as i32, y as i32, 0, 0);
            self.popover.set_pointing_to(Some(&position));
            self.popover.popup();
        }

        #[template_callback]
        fn on_row_selected(&self, row: Option<&ListBoxRow>) {
            let obj = self.obj();
            let selected = row.is_some();

            if selected {
                obj.emit_by_name::<()>("selection-changed", &[&*obj, &GSidebarSelection::all()]);
            }
        }

        #[template_callback]
        fn format_item_count(&self, count: u32) -> String {
            count.to_string()
        }

        #[template_callback]
        fn is_greater_than_zero(&self, count: u32) -> bool {
            count > 0
        }
    }
}

glib::wrapper! {
    pub struct AllArticlesRow(ObjectSubclass<imp::AllArticlesRow>)
        @extends Widget, Bin;
}

impl Default for AllArticlesRow {
    fn default() -> Self {
        glib::Object::new::<Self>()
    }
}

impl AllArticlesRow {
    pub fn clear_selection(&self) {
        self.imp().list.unselect_all();
    }

    pub fn select(&self) {
        let imp = self.imp();
        imp.list.select_row(Some(&*imp.row));
    }

    pub fn activate(&self) {
        self.emit_by_name::<()>("activated", &[self]);
    }

    pub fn connect_activated<F: Fn(&Self) + 'static>(&self, f: F) -> SignalHandlerId {
        self.connect_local("activated", false, move |args| {
            let all_articles = args[1]
                .get::<Self>()
                .expect("The value needs to be of type `AllArticlesRow`");
            f(&all_articles);
            None
        })
    }
}
